#ifndef ILLUMESTIMATORS_BAYESIANESTIMATOR_H
#define ILLUMESTIMATORS_BAYESIANESTIMATOR_H

#include "superpixelsegmentation.h"
#include "illuminantestimator.h"
#include <vector>

namespace illumestimators {

/** The BayesianEstimator class provides a Bayesian illuminant estimator.
 * Read C. Rosenberg, T. Minka, and A. Ladsariya, “Bayesian color constancy with non-gaussian models,”
 * Advances in Neural Information Processing Systems, vol. 16, 2004.
 */
class BayesianEstimator : public IlluminantEstimator
{
	struct ltvec3d
	{
		bool operator()(const cv::Vec3d &a, const cv::Vec3d &b) const
		{
			double diff = a[0] - b[0];
			if (fabs(diff) <= std::numeric_limits<double>::epsilon()) {
				diff = a[1] - b[1];
				if (fabs(diff) <= std::numeric_limits<double>::epsilon()) {
					diff = a[2] - b[2];
					return diff <= 0;
				} else
					return diff < 0;
			} else
				return diff < 0;
		}
	};


public:
	enum Method { POSTERIOR_MEAN, MIN_AVERAGE_LOSS_ANGULAR, MIN_AVERAGE_LOSS_EUCLIDEAN };

public:
	BayesianEstimator(Method method = POSTERIOR_MEAN, int verbosity = 0, bool cache = false);
	BayesianEstimator(Method method, const std::vector<cv::Vec3d>& illuminants, const cv::Mat_<double>& logReflectancePrior, int verbosity = 0, bool cache = false);
	~BayesianEstimator();

public:
	std::string name() const;
	std::string identifier() const;

public:
	cv::Mat_<double> getLogReflectancePrior() { return m_logReflectancePrior; }

public:
	using IlluminantEstimator::estimateIlluminant;
	Illum estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask) const;
	Illum estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const superpixels::Superpixel &superpixel, const cv::Mat_<unsigned char>& mask) const;

	// EXTENSION FOR CPCV
	void estimateIlluminants(
		const cv::Mat_<cv::Vec3d>& image,
		const std::vector<superpixels::Superpixel>& superpixel,
		std::vector<Illum> &illuminants,
		const cv::Mat_<unsigned char>& mask
	);
	Illum perform_actual_estimation(const std::vector<cv::Vec3d> &illuminants, const std::vector<double> &likelihoods) const;

public:
	bool train(const std::vector<std::string>& imageFiles, const std::vector<std::string>& colorspaces, const std::vector<cv::Vec3d>& illuminants, const std::vector<std::string>& maskFiles);
	bool train(const std::vector<cv::Mat_<int> > histograms, const std::vector<cv::Vec3d>& illuminants);


	// EXTENSIONS FOR CPCV
	bool addTrainingData(
		cv::Mat_<cv::Vec3d> img,
		const std::vector<superpixels::Superpixel>& superpixels,
		const std::vector<cv::Vec3d>& illuminants,
		cv::Mat_<unsigned char> &mask
	);

	bool finalizeTraining();

	cv::Mat_<int> calculateReflectancesHistogram(
		const cv::Mat_<cv::Vec3d> &img,
		const superpixels::Superpixel &superpixel,
		const cv::Vec3d& illuminant,
		const cv::Mat_<unsigned char> &mask
	);

public:
	bool save(const std::string& filename) const;
	bool load(const std::string& filename);
	virtual int error(); // dummy implementation, returns always 0

public:
	cv::Mat_<int> calculateReflectancesHistogram(const std::string& imageFile, const std::string& colorspace, const cv::Vec3d& illuminant, const std::string& maskFile);

private:
	double sumlog(const std::vector<double>& values) const;
	cv::Mat_<int> calculateHistogram(const std::vector<cv::Vec3d>& pixels) const;
	std::vector<cv::Vec3d> calculateReflectances(const std::vector<cv::Vec3d>& pixels, const cv::Vec3d& illuminant) const;
	double logLikelihoodRosenberg(const std::vector<cv::Vec3d>& pixels, const cv::Vec3d& illuminant) const;

	// CHANGE PRIVATE->PUBLIC FOR CPCV
public:
	virtual void preprocessImage(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask, cv::Mat_<cv::Vec3d>& outputImage, cv::Mat_<unsigned char>& outputMask) const;

private:
	Method m_method;
	std::vector<cv::Vec3d> m_illuminants;
	cv::Mat_<double> m_logReflectancePrior;
	int m_size;
	int verbosity;
	bool m_cache;

	// INSERTED
	std::set<cv::Vec3d, ltvec3d> distinct_illums;
};

} // namespace illumestimators

#endif // ILLUMESTIMATORS_BAYESIANESTIMATOR_H
