#include "gamutmapping2destimator.h"
#include "evaluation/error.h"
#include "common/color.h"
#include "common/statistics.h"
#include "common/derivative.h"
#include "common/mask.h"
#include "common/scale.h"
#include "common/convexpolygon.h"
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <iostream>
#include <ctime>
#include <limits>

namespace illumestimators {

GamutMapping2DEstimator::GamutMapping2DEstimator(Method method, int n, double sigma, int verbosity, bool cache) :
	m_method(method),
	m_n(n),
	m_sigma(sigma),
	m_verbosity(verbosity),
	m_cache(cache)
{
	std::vector<cv::Point2f> canonical;
	canonical.push_back(cv::Point2f(0, 0));
	canonical.push_back(cv::Point2f(0, 1));
	canonical.push_back(cv::Point2f(1, 0));
	canonical.push_back(cv::Point2f(1, 1));
	srand(time(NULL));
	setCanonicalGamut(canonical);
}

GamutMapping2DEstimator::GamutMapping2DEstimator(const std::vector<cv::Point2f>& canonical, Method method, int n, double sigma, int verbosity, bool cache) :
	m_method(method),
	m_n(n),
	m_sigma(sigma),
	m_verbosity(verbosity),
	m_cache(cache)
{
	setCanonicalGamut(canonical);
}

GamutMapping2DEstimator::~GamutMapping2DEstimator()
{

}

std::string GamutMapping2DEstimator::name() const
{
	std::stringstream name;
	name << "GamutMapping2DEstimator(method = " << m_method << ", n = " << m_n << ", sigma = " << m_sigma << ")";

	return name.str();
}

std::string GamutMapping2DEstimator::identifier() const
{
	std::stringstream identifier;
	identifier << "gamut2d-" << m_n << "-" << m_sigma;

	return identifier.str();
}

void GamutMapping2DEstimator::setCanonicalGamut(const std::vector<cv::Point2f>& canonical)
{
	std::vector<cv::Point2f> points = canonical;
	points.push_back(cv::Point2f(0, 0));

	m_canonical.clear();

	cv::convexHull(cv::Mat(points), m_canonical, false);

	if (m_verbosity > 1) {
		std::cout << "canonical gamut:" << std::endl;
		for (unsigned int i = 0; i < m_canonical.size(); i++) {
			std::cout << m_canonical[i].x << " " << m_canonical[i].y << std::endl;
		}
		std::cout << std::endl;
	}
}

void GamutMapping2DEstimator::preprocessImage(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask, cv::Mat_<cv::Vec3d>& outputImage, cv::Mat_<unsigned char>& outputMask) const
{
	outputImage = image.clone();
	if ((image.rows != mask.rows) || (image.cols != mask.cols)) {
		std::cerr << "No mask!" << std::endl;
		outputMask = cv::Mat_<unsigned char>::zeros(image.rows, image.cols);
	} else {
		outputMask = mask.clone();
	}

	Mask::maskSaturatedPixels(outputImage, outputMask, 254.0 / 255);
	Mask::maskMinPixels(outputImage, outputMask, 1.0 / 255);

	if (m_sigma > 0) {
		if (m_n == 0) {
			cv::GaussianBlur(outputImage, outputImage, cv::Size(0, 0), m_sigma, m_sigma);
		} else if (m_n > 0) {
			outputImage = Derivative::normDerivativeFilter(outputImage, m_n, m_sigma);
		}
	}
}

double GamutMapping2DEstimator::sumlog(const std::vector<double>& values) const
{
	const double max = Statistics::max(values);
	double sum = 0;

	for (unsigned int i = 0; i < values.size(); i++) {
		sum += exp(values[i] - max);
	}

	return max + log(sum);
}

Illum GamutMapping2DEstimator::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask) const
{
	cv::Mat_<cv::Vec3d> inputImage;
	cv::Mat_<unsigned char> inputMask;

	preprocessImage(image, mask, inputImage, inputMask);

	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(inputImage, inputMask);

	std::vector<cv::Point2f> chromaticities;

	for (unsigned int i = 0; i < pixels.size(); i++) {
		if (pixels[i][0] > 0) {
			chromaticities.push_back(cv::Point2f(pixels[i][2] / pixels[i][0], pixels[i][1] / pixels[i][0]));
		}
	}

	chromaticities.push_back(cv::Point2f(0, 0));

	std::vector<cv::Point2f> imageGamut = calculateImageGamut(chromaticities);

	std::vector<cv::Point2f> possibleMappings = calculatePossibleMappings(imageGamut, m_canonical);
	std::vector<cv::Point2f> possibleIlluminants = calculateSetOfPossibleIlluminants(m_illuminants);

	std::vector<cv::Vec3d> randomVectors = calculateMonteCarloIntersection(possibleMappings, possibleIlluminants);

	return findBestGamut(randomVectors);
}

Illum GamutMapping2DEstimator::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const superpixels::Superpixel &superpixel, const cv::Mat_<unsigned char>& mask) const
{
	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(image, superpixel, mask);

	std::vector<cv::Point2f> chromaticities;

	for (unsigned int i = 0; i < pixels.size(); i++) {
		if (pixels[i][0] > std::numeric_limits<double>::epsilon()) {
			chromaticities.push_back(cv::Point2f(pixels[i][2] / pixels[i][0], pixels[i][1] / pixels[i][0]));
		}
	}

	chromaticities.push_back(cv::Point2f(0, 0));

	std::vector<cv::Point2f> imageGamut = calculateImageGamut(chromaticities);

	std::vector<cv::Point2f> possibleMappings = calculatePossibleMappings(imageGamut, m_canonical);
	std::vector<cv::Point2f> possibleIlluminants = calculateSetOfPossibleIlluminants(m_illuminants);

	std::vector<cv::Vec3d> randomVectors = calculateMonteCarloIntersection(possibleMappings, possibleIlluminants);

	return findBestGamut(randomVectors);
}

void GamutMapping2DEstimator::estimateIlluminants(
	const cv::Mat_<cv::Vec3d>& image,
	const std::vector<superpixels::Superpixel>& superpixels,
	std::vector<Illum> &estimated_illuminants,
	const cv::Mat_<unsigned char>& mask) const
{
	estimated_illuminants = std::vector<Illum>(superpixels.size(), Illum());

	for (unsigned int s = 0; s < superpixels.size(); ++s) {

		cv::Vec3d estimate = cv::Vec3d(0, 0, 0);

		const std::vector<cv::Point> &coord = superpixels[s].coordinates;
		cv::vector<cv::Vec3d> pixels(coord.size());
		int count = 0;
		for (int i = 0; i < (int)coord.size(); ++i) {
			if (mask[coord[i].y][coord[i].x] > 0) continue;
			pixels[i] = image[ coord[i].y ][ coord[i].x ];
			count++;
		}
		pixels.resize(count);

		std::vector<cv::Point2f> chromaticities;

		for (unsigned int i = 0; i < pixels.size(); i++) {
			if (pixels[i][0] > 0) {
				chromaticities.push_back(cv::Point2f(pixels[i][2] / pixels[i][0], pixels[i][1] / pixels[i][0]));
			}
		}

		chromaticities.push_back(cv::Point2f(0, 0));

		std::vector<cv::Point2f> imageGamut = calculateImageGamut(chromaticities);

		std::vector<cv::Point2f> possibleMappings = calculatePossibleMappings(imageGamut, m_canonical);
		std::vector<cv::Point2f> possibleIlluminants = calculateSetOfPossibleIlluminants(m_illuminants);

		std::vector<cv::Vec3d> randomVectors = calculateMonteCarloIntersection(possibleMappings, possibleIlluminants);

		estimated_illuminants[s] = findBestGamut(randomVectors);
	}
}

std::vector<cv::Point2f> GamutMapping2DEstimator::calculateCanonicalGamut(const std::string& filename, const std::string& colorspace, const cv::Vec3d& illuminant, const cv::Mat_<unsigned char>& mask, bool preprocess)
{
	cv::Mat_<cv::Vec3d> image = cleanlyReadImage(filename, colorspace);
	if (m_verbosity > 0) {
		std::cout << "min=" << Statistics::min(Statistics::min(image)) << ", max=" << Statistics::max(Statistics::max(image)) << std::endl << std::endl;
	}

	cv::Mat_<cv::Vec3d> inputImage;
	cv::Mat_<unsigned char> inputMask;
	if (preprocess) preprocessImage(image, mask, inputImage, inputMask);
	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(inputImage, inputMask);

	for (unsigned int i = 0; i < pixels.size(); i++) {
		for (int c = 0; c < 3; c++) {
			pixels[i][c] /= illuminant[c];
		}
	}

	Scale::scaleByMax(pixels);

	std::vector<cv::Point2f> chromaticities;

	for (unsigned int i = 0; i < pixels.size(); i++) {
		if ((pixels[i][0] >= (1.0 / 255)) && (pixels[i][1] >= (1.0 / 255)) && (pixels[i][2] >= (1.0 / 255))) {
			chromaticities.push_back(cv::Point2f(pixels[i][2] / pixels[i][0], pixels[i][1] / pixels[i][0]));
		}
	}

	chromaticities.push_back(cv::Point2f(0, 0));
	std::vector<cv::Point2f> hull;
	cv::convexHull(cv::Mat(chromaticities), hull, false);

	return hull;
}

std::vector<cv::Point2f> GamutMapping2DEstimator::calculateCanonicalGamutSuperpixel(
	cv::Mat_<cv::Vec3d> &img,
	const superpixels::Superpixel &superpixel,
	const cv::Vec3d& illuminant,
	const cv::Mat_<unsigned char>& mask)
{
	const std::vector<cv::Point> &coord = superpixel.coordinates;
	//std::vector<cv::Vec3d> pixels; = Mask::unmaskedPixels(inputImage, inputMask);

	cv::vector<cv::Vec3d> pixels(coord.size());
	int count = 0;
	for (int i = 0; i < (int)coord.size(); ++i) {
		if (mask[coord[i].y][coord[i].x] > 0) continue;
		pixels[i] = img[ coord[i].y ][ coord[i].x ];
		count++;
	}
	pixels.resize(count);

	for (unsigned int i = 0; i < pixels.size(); i++) {
		for (int c = 0; c < 3; c++) {
			pixels[i][c] /= illuminant[c];
		}
	}

	Scale::scaleByMax(pixels);

	std::vector<cv::Point2f> chromaticities;

	for (unsigned int i = 0; i < pixels.size(); i++) {
		if ((pixels[i][0] >= (1.0 / 255)) && (pixels[i][1] >= (1.0 / 255)) && (pixels[i][2] >= (1.0 / 255))) {
			chromaticities.push_back(cv::Point2f(pixels[i][2] / pixels[i][0], pixels[i][1] / pixels[i][0]));
		}
	}

	std::vector<cv::Point2f> hull;
	if (chromaticities.size() < 4) return hull; // failure
	cv::convexHull(cv::Mat(chromaticities), hull, false);

	return hull;
}


// EXTENSIONS FOR CPCV
bool GamutMapping2DEstimator::addTrainingData(
	cv::Mat_<cv::Vec3d> &img,
	const std::vector<superpixels::Superpixel>& superpixels,
	const std::vector<cv::Vec3d>& illuminants,
	cv::Mat_<unsigned char> &mask
)
{
	for (int i = 0; i < (int)superpixels.size(); ++i) {
		std::vector<cv::Point2f> gamut = calculateCanonicalGamutSuperpixel(img, superpixels[i], illuminants[i], mask);
		if (gamut.size() > 0) {
			m_canonical.insert(m_canonical.end(), gamut.begin(), gamut.end());
		}
	}
	return true;
}

bool GamutMapping2DEstimator::finalizeTraining()
{
	std::vector<cv::Point2f> tmp_canonical(m_canonical);
	setCanonicalGamut(m_canonical);
	return true;
}


bool GamutMapping2DEstimator::train(const std::vector<std::string>& imageFiles, const std::vector<std::string>& colorspaces, const std::vector<cv::Vec3d>& illuminants, const std::vector<std::string>& maskFiles)
{
	if ((imageFiles.size() != colorspaces.size()) || (imageFiles.size() != illuminants.size())) {
		return false;
	}

	if (imageFiles.size() < 1) {
		return false;
	}

	m_illuminants = illuminants;

	std::vector<cv::Point2f> canonical;

	for (unsigned int i = 0; i < imageFiles.size(); i++) {
		if (m_cache) {
			static std::map<std::string, std::vector<cv::Point2f> > gamutCache;

			// TODO: Resolve proper path, load/save
			if (gamutCache.find(imageFiles[i]) != gamutCache.end()) {
				canonical.insert(canonical.end(), gamutCache[imageFiles[i]].begin(), gamutCache[imageFiles[i]].end());
			} else {
				cv::Mat_<unsigned char> mask = cv::imread(maskFiles[i], 0);
				std::vector<cv::Point2f> gamut = calculateCanonicalGamut(imageFiles[i], colorspaces[i], illuminants[i], mask, true);
				gamutCache[imageFiles[i]] = gamut;
				canonical.insert(canonical.end(), gamut.begin(), gamut.end());
			}
		} else {
			cv::Mat_<unsigned char> mask = cv::imread(maskFiles[i], 0);
			std::vector<cv::Point2f> gamut = calculateCanonicalGamut(imageFiles[i], colorspaces[i], illuminants[i], mask, true);
			canonical.insert(canonical.end(), gamut.begin(), gamut.end());
		}
	}

	setCanonicalGamut(canonical);

	return true;
}

bool GamutMapping2DEstimator::train(const std::vector<std::vector<cv::Point2f> >& gamuts, const std::vector<cv::Vec3d>& illuminants)
{
	if (gamuts.size() < 1) {
		return false;
	}

	m_illuminants = illuminants;

	std::vector<cv::Point2f> canonical;

	for (unsigned int i = 0; i < gamuts.size(); i++) {
		canonical.insert(canonical.end(), gamuts[i].begin(), gamuts[i].end());
	}

	setCanonicalGamut(canonical);

	return true;
}

std::vector<cv::Point2f> GamutMapping2DEstimator::calculateImageGamut(const std::vector<cv::Point2f>& chromaticities) const
{
	std::vector<cv::Point2f> imageGamut;

	cv::convexHull(cv::Mat(chromaticities), imageGamut, false);

	if (m_verbosity > 0) {
		std::cout << "image gamut:" << std::endl;
		for (unsigned int i = 0; i < imageGamut.size(); i++) {
			std::cout << imageGamut[i].x << " " << imageGamut[i].y << std::endl;
		}
		std::cout << std::endl;
	}

	return imageGamut;
}

std::vector<cv::Point2f> GamutMapping2DEstimator::calculatePossibleMappings(const std::vector<cv::Point2f>& imageGamut, const std::vector<cv::Point2f>& canonicalGamut) const
{
	std::vector<ConvexPolygon> mappingsVector;

	for (unsigned int i = 0; i < imageGamut.size(); i++) {
		ConvexPolygon mappings;

		for (unsigned int j = 0; j < canonicalGamut.size(); j++) {
			double r = imageGamut[i].x;
			double g = imageGamut[i].y;

			if (r < 0.001) {
				r = 0.001;
			}
			if (g < 0.001) {
				g = 0.001;
			}

			mappings.addPoint(cv::Point2f(canonicalGamut[j].x / r, canonicalGamut[j].y / g));
		}

		mappingsVector.push_back(mappings);
	}

	if (mappingsVector.size() < 1) {
		std::cerr << "Failed to calculate set of possible mappings." << std::endl;
		return std::vector<cv::Point2f>();
	}

	ConvexPolygon possibleMappings = mappingsVector[0];

	for (unsigned int i = 0; i < mappingsVector.size(); i++) {
		if (m_verbosity > 1) {
			std::cout << "mapping " << i << ":" << std::endl;
			for (unsigned int j = 0; j < mappingsVector[i].vertices().size(); j++) {
				std::cout << mappingsVector[i].vertices()[j].x << " " << mappingsVector[i].vertices()[j].y << std::endl;
			}
			std::cout << std::endl;
		}

		possibleMappings = possibleMappings.intersect(mappingsVector[i]);

		if (m_verbosity > 1) {
			std::cout << "intersection " << i << ":" << std::endl;
			for (unsigned int j = 0; j < possibleMappings.vertices().size(); j++) {
				std::cout << possibleMappings.vertices()[j].x << " " << possibleMappings.vertices()[j].y << std::endl;
			}
			std::cout << std::endl;
		}

		if (possibleMappings.vertices().size() < 3) {
			std::cerr << "Failed to calculate set of possible mappings." << std::endl;
			return std::vector<cv::Point2f>();
		}
	}

	if (m_verbosity > 0) {
		std::cout << "possible mappings:" << std::endl;
		for (unsigned int i = 0; i < possibleMappings.vertices().size(); i++) {
			std::cout << possibleMappings.vertices()[i].x << " " << possibleMappings.vertices()[i].y << std::endl;
		}
		std::cout << std::endl;
	}

	return possibleMappings.vertices();
}

std::vector<cv::Point2f> GamutMapping2DEstimator::calculateSetOfPossibleIlluminants(const std::vector<cv::Vec3d>& illuminants) const
{
	std::vector<cv::Point2f> illuminantChromaticities;

	for (unsigned int i = 0; i < illuminants.size(); i++) {
		illuminantChromaticities.push_back(cv::Point2f(illuminants[i][2], illuminants[i][1]));

		for (double dx = 0.95; dx <= 1.05; dx += 0.001) {
			for (double dy = 0.95; dy <= 1.05; dy += 0.001) {
				illuminantChromaticities.push_back(cv::Point2f(illuminants[i][2] * dx, illuminants[i][1] * dy));
			}
		}
	}

	std::vector<cv::Point2f> illuminantHull;

	cv::convexHull(cv::Mat(illuminantChromaticities), illuminantHull, false);

	if (m_verbosity > 0) {
		std::cout << "possible illuminants:" << std::endl;
		for (unsigned int i = 0; i < illuminantHull.size(); i++) {
			std::cout << illuminantHull[i].x << " " << illuminantHull[i].y << std::endl;
		}
		std::cout << std::endl;
	}

	return illuminantHull;
}

std::vector<cv::Vec3d> GamutMapping2DEstimator::calculateMonteCarloIntersection(const std::vector<cv::Point2f>& possibleMappings, const std::vector<cv::Point2f>& possibleIlluminants) const
{
	std::vector<cv::Vec3d> randomVectors;

	if (possibleMappings.size() < 3) {
		std::cerr << "Failed to calculate Monte Carlo intersection." << std::endl;
		randomVectors.insert(randomVectors.end(), m_illuminants.begin(), m_illuminants.end());
		return randomVectors;
	}

	const unsigned int randomVectorsNumber= 1000;
	const unsigned int maxIterations = 100000;

	unsigned int counter = 0;

	while (randomVectors.size() < randomVectorsNumber) {
		cv::Vec3d randomVector;

		for (unsigned int i = 0; i < 3; i++) {
			randomVector[i] = ((rand() % 999)+1) / 1000.0;
		}

		const cv::Point2f illuminant(randomVector[2] / (randomVector[0] + randomVector[1] + randomVector[2]), randomVector[1] / (randomVector[0] + randomVector[1] + randomVector[2]));
		const cv::Point2f mapping(1 / (randomVector[2] / randomVector[0]), 1 / (randomVector[1] / randomVector[0]));

		const bool inPossibleMappings = cv::pointPolygonTest(cv::Mat(possibleMappings), mapping, false) >= 0;
		const bool inPossibleIlluminants = (possibleIlluminants.size() < 3) || (cv::pointPolygonTest(cv::Mat(possibleIlluminants), illuminant, false) >= 0);

		if (inPossibleMappings && inPossibleIlluminants) {
			randomVectors.push_back(randomVector);
		}

		if (counter >= maxIterations) {
			std::cerr << "Reached maximum number of iterations." << std::endl;

			if (randomVectors.size() < 1) {
				randomVectors.insert(randomVectors.end(), m_illuminants.begin(), m_illuminants.end());
				return randomVectors;
			}

			break;
		}

		counter++;
	}

	if (m_verbosity > 1) {
		std::cout << "random vectors:" << std::endl;
		for (unsigned int i = 0; i < randomVectors.size(); i++) {
			std::cout << randomVectors[i][2] << " " << randomVectors[i][1] << " " << randomVectors[i][0] << std::endl;
		}
		std::cout << std::endl;
	}

	return randomVectors;
}

Illum GamutMapping2DEstimator::findBestGamut(std::vector<cv::Vec3d>& randomVectors) const
{
	Illum estimate(true);

	if (m_method == MAX) {
		double maxtrace = -std::numeric_limits<double>::max();

		for (unsigned int i = 0; i < randomVectors.size(); i++) {
			const double alpha = 1 / (randomVectors[i][2] / randomVectors[i][0]);
			const double beta = 1 / (randomVectors[i][1] / randomVectors[i][0]);
			const double trace = alpha + beta;

			if (trace > maxtrace) {
				maxtrace = trace;
				for (int c = 0; c < 3; ++c) estimate[c] = randomVectors[i][c];
			}
		}
	} else if (m_method == MEDIAN) {
		double mindist = std::numeric_limits<double>::max();

		for (unsigned int i = 0; i < randomVectors.size(); i++) {
			double dist = 0;

			for (unsigned int j = 0; j < randomVectors.size(); j++) {
				if (i == j) {
					continue;
				}

				const double error = Error(randomVectors[i], randomVectors[j]).toAngularErrorRad();

				if (error > dist) {
					dist = error;
				}
			}

			if (dist < mindist) {
				mindist = dist;
				for (int c = 0; c < 3; ++c) estimate[c] = randomVectors[i][c];
			}
		}
	} else if (m_method == MEAN) {
		cv::Vec3d the_mean = Statistics::mean(randomVectors);
		for (int c = 0; c < 3; ++c) estimate[c] = the_mean[c];
	}

	estimate = Color::calculateChromaticities(estimate);

	return estimate;
}

bool GamutMapping2DEstimator::save(const std::string& filename) const
{
	cv::FileStorage fs(filename, cv::FileStorage::WRITE);

	if (!fs.isOpened()) {
		std::cerr << "unable to open file storage " << filename << std::endl;
		return false;
	}

	fs << "name" << "GamutMapping2DEstimator";
	fs << "method" << m_method;
	fs << "n" << m_n;
	fs << "sigma" << m_sigma;

	fs << "canonical" << "[";
	for (unsigned int i = 0; i < m_canonical.size(); i++) {
		fs << m_canonical[i];
	}
	fs << "]";

	fs << "illuminants" << "[";
	for (unsigned int i = 0; i < m_illuminants.size(); i++) {
		fs << m_illuminants[i];
	}
	fs << "]";

	return true;
}

bool GamutMapping2DEstimator::load(const std::string& filename)
{
	cv::FileStorage fs(filename, cv::FileStorage::READ);

	if (!fs.isOpened()) {
		std::cerr << "unable to open file storage " << filename << std::endl;
		return false;
	}

	std::string name = fs["name"];

	if (name != "GamutMapping2DEstimator") {
		std::cerr << "wrong method name: expected GamutMapping2DEstimator, obtained " << name << std::endl;
		return false;
	}

	m_method = (Method) (int) fs["method"];
	m_n = fs["n"];
	m_sigma = fs["sigma"];

	m_canonical.clear();
	cv::FileNode canonical = fs["canonical"];
	if (canonical.type() == cv::FileNode::SEQ) {
		for (unsigned int i = 0; i < canonical.size(); i++) {
			CV_Assert(canonical[i].type() == cv::FileNode::SEQ && canonical[i].size() == 2);
			m_canonical.push_back(cv::Point2f((double) canonical[i][0], (double) canonical[i][1]));
		}
	}

	m_illuminants.clear();
	cv::FileNode illuminants = fs["illuminants"];
	if (illuminants.type() == cv::FileNode::SEQ) {
		for (unsigned int i = 0; i < illuminants.size(); i++) {
			CV_Assert(illuminants[i].type() == cv::FileNode::SEQ && illuminants[i].size() == 3);
			m_illuminants.push_back(cv::Vec3d((double) illuminants[i][0], (double) illuminants[i][1], (double) illuminants[i][2]));
		}
	}

	return true;
}

int GamutMapping2DEstimator::error() {
	return 0;
}


} // namespace illumestimators
